#pragma once

// version that maintains the graph simple using a set data structure
// multiedgelist.hpp allows multiple edges using list instead of set

#include <fstream>
// #include <set>
#include <boost/unordered_set.hpp>
#include <vector>

using namespace std;

/* edge list representation */
struct  Edge {
  int tail;
  int head;
  // cap is mutable in order to be changed inplace in a set<Edge>
  // the < operator cannot use cap.
  mutable long cap; 

  bool operator==(const Edge& other) const {
    return (this->tail == other.tail && this->head == other.head) || 
      (this->tail == other.head && this->head == other.tail);
  } 

  friend std::size_t hash_value(Edge const& e);

  /*  bool operator!=(const Edge& other) const {
     return !(*this == other); 
    } */ 
  /* bool operator<(const Edge& rhs) const {
      return ((*this).tail < rhs.tail) || ((*this).tail == rhs.tail && (*this).head < rhs.head); }
      bool operator>(const Edge& rhs) const {
    if (*this == rhs)
      return false;
      return ((*this).tail > rhs.tail) || ((*this).tail == rhs.tail && (*this).head > rhs.tail); } */
};

class Edgelist {
public:
  typedef boost::unordered_set<Edge> edges_t;
  typedef boost::unordered_set<Edge>::iterator edge_iterator;
  // typedef set<Edge> edges_t;
  // typedef std::set<Edge>::iterator edge_iterator;
private:
  int number_of_nodes, 
      number_of_edges;    // number of vertices/edges 
  int source, sink; // source/sink if it is a max flow instance
  //  set<Edge> edges;  
  edges_t edges;

 public:  
  vector<int> degrees;
  Edgelist(int n, int m);
  Edgelist();
  //  ~Edgelist();
  
  Edge add_edge(int tail, int head, long cap);
  int read_dimacs (ifstream& fp);
  //  void print();
  int num_nodes() { return number_of_nodes; }
  int num_edges() { return number_of_edges; }
  void set_size(int n);
  int get_source() { return source; }
  int get_sink() { return sink; }
  void set_source(int src) { source = src; }
  void set_sink(int snk) { sink = snk; }
  void clear();

  edge_iterator begin() { return edges.begin(); }
  edge_iterator end() { return edges.end(); }
  friend ostream& operator<<(ostream& os, Edgelist& el);

};

struct Simple_Edge {
  int tail;
  int head;
  long cap; 
};

class Edgelist_vector {
  int number_of_nodes, 
      number_of_edges;    // number of vertices/edges 
  int source, sink; // source/sink if it is a max flow instance
  //  set<Edge> edges;  
  vector<Simple_Edge> edges;

 public:  
  vector<int> degrees;
  Edgelist_vector(int n);
  Edgelist_vector();
  //  ~Edgelist();
  
  Simple_Edge add_edge(int tail, int head, long cap);
  int read_dimacs (ifstream& fp);
  //  void print();
  int num_nodes() { return number_of_nodes; }
  int num_edges() { return number_of_edges; }
  void set_size(int n);
  int get_source() { return source; }
  int get_sink() { return sink; }
  void set_source(int src) { source = src; }
  void set_sink(int snk) { sink = snk; }
  void clear();

  // typedef std::set<Edge>::iterator edge_iterator;
  vector<Simple_Edge>::iterator begin() { return edges.begin(); }
  vector<Simple_Edge>::iterator end() { return edges.end(); }
  friend ostream& operator<<(ostream& os, Edgelist_vector& el);

  typedef vector<Simple_Edge> edges_t;
  typedef vector<Simple_Edge>::iterator edge_iterator;
};
