#include "task.hpp"
#include <mpi.h>
#include <stdlib.h>
#include <iostream>

using namespace std;

Task::Task (int task_number, int s, int t, int pivot_size, int part_len, int refine)
{
  this->task_number = task_number;
  this->s = s;
  this->t = t;
  this->pivot_size = pivot_size;
  this->part_len = part_len;
  this->refine = refine;
}

Task::Task()
{
}

Task::~Task()
{
}

Response::Response(int n)
{
  cut = new int[n];
}

Response::~Response()
{
  delete[] cut;
}

// functions from C implementation ... 
// can be transformed into C++ methods

int* task_buffer;
int* reply_buffer;
int  task_buffer_size;
int  reply_buffer_size;

void print_node_map(int *node_map, int n, ostream& os)
{
  int i;

  os << "c Nodemap:" << endl;
  for (i=0; i<n; i++)
    os << "c " <<  i+1 << " --> " <<  node_map[i] << endl;
}

int sendTask(Task& task, int dest, int N, int* node_map)
{
  int position = 0; //current position in the buffer
  
  MPI_Pack(&task.task_number, 1, MPI_INT, task_buffer, task_buffer_size, &position, MPI_COMM_WORLD);
  MPI_Pack(&task.s, 1, MPI_INT, task_buffer, task_buffer_size, &position, MPI_COMM_WORLD);
  MPI_Pack(&task.t, 1, MPI_INT, task_buffer, task_buffer_size, &position, MPI_COMM_WORLD);
  MPI_Pack(&task.pivot_size, 1, MPI_INT, task_buffer, task_buffer_size, &position, MPI_COMM_WORLD);
  MPI_Pack(&task.part_len, 1, MPI_INT, task_buffer, task_buffer_size, &position, MPI_COMM_WORLD);
  MPI_Pack(&task.refine, 1, MPI_INT, task_buffer, task_buffer_size, &position, MPI_COMM_WORLD);
  MPI_Pack(node_map, N, MPI_INT, task_buffer, task_buffer_size, &position, MPI_COMM_WORLD);

  int r =  MPI_Send(task_buffer, position, MPI_PACKED, dest, 1, MPI_COMM_WORLD);
  return r;
}

void print_task(Task& task, int* node_map, int N)
{
  cerr << "c Number: " << task.task_number << endl;
  cerr << "c s/t:    " << task.s << " / " << task.t << endl;
  cerr << "c Pivot size: " << task.pivot_size << endl;
  cerr << "c part_len: " << task.part_len << endl;
  // print_node_map(node_map, N, cerr);
}

void print_reply (Response& msg, int N)
{
  cerr << "c    Task Number: " << msg.task_number << " s/t: " << msg.s+1 << " / " 
       << msg.t+1 << endl;
  cerr << "c    Flow value: " << msg.flow << endl;
  cerr << "c    Cut: ";
  for (int i=0; i<N; i++)
    if (msg.cut[i])
      cerr << i+1 << " ";
  cerr << endl;
}

int receiveResponse(Response& msg, int N) {
  int position = 0;
  int tag = 1;
  MPI_Status status;

  // receive the structure in the buffer
  MPI_Recv(reply_buffer, reply_buffer_size, MPI_PACKED, MPI_ANY_SOURCE, tag, MPI_COMM_WORLD, &status);
//  printf("recebeu...%d\n",r);

  // obtain the structure fields from the buffer
  MPI_Unpack(reply_buffer, reply_buffer_size, &position, &msg.task_number, 1, MPI_INT, MPI_COMM_WORLD);
  MPI_Unpack(reply_buffer, reply_buffer_size, &position, &msg.s, 1, MPI_INT, MPI_COMM_WORLD);
  MPI_Unpack(reply_buffer, reply_buffer_size, &position, &msg.t, 1, MPI_INT, MPI_COMM_WORLD);
  MPI_Unpack(reply_buffer, reply_buffer_size, &position, &msg.graph_size, 1, MPI_INT, MPI_COMM_WORLD);
  MPI_Unpack(reply_buffer, reply_buffer_size, &position, &msg.flow, 1, MPI_DOUBLE, MPI_COMM_WORLD);
  MPI_Unpack(reply_buffer, reply_buffer_size, &position, &msg.time_per_mincut, 1, MPI_DOUBLE, MPI_COMM_WORLD);
  MPI_Unpack(reply_buffer, reply_buffer_size, &position, &msg.time_to_contract, 1, MPI_DOUBLE, MPI_COMM_WORLD);
  MPI_Unpack(reply_buffer, reply_buffer_size, &position, &msg.num_of_cuts, 1, MPI_INT, MPI_COMM_WORLD);
  MPI_Unpack(reply_buffer, reply_buffer_size, &position, &msg.balance_factor, 1, MPI_FLOAT, MPI_COMM_WORLD);

  MPI_Unpack(reply_buffer, reply_buffer_size, &position, msg.cut, N, MPI_INT, MPI_COMM_WORLD);

  return status.MPI_SOURCE;
}

int receiveTask(Task& task, int N, int *node_map) {
  int position = 0;
  int tag = 1;
  MPI_Status status;

  // receive the structure in the buffer
  MPI_Recv(task_buffer, task_buffer_size, MPI_PACKED, MPI_ANY_SOURCE, tag, MPI_COMM_WORLD, &status);
//  printf("recebeu...%d\n",r);

  // obtain the structure fields from the buffer
  MPI_Unpack(task_buffer, task_buffer_size, &position, &task.task_number, 1, MPI_INT, MPI_COMM_WORLD);
  MPI_Unpack(task_buffer, task_buffer_size, &position, &task.s, 1, MPI_INT, MPI_COMM_WORLD);
  MPI_Unpack(task_buffer, task_buffer_size, &position, &task.t, 1, MPI_INT, MPI_COMM_WORLD);
  MPI_Unpack(task_buffer, task_buffer_size, &position, &task.pivot_size, 1, MPI_INT, MPI_COMM_WORLD);
  MPI_Unpack(task_buffer, task_buffer_size, &position, &task.part_len, 1, MPI_INT, MPI_COMM_WORLD);
  MPI_Unpack(task_buffer, task_buffer_size, &position, &task.refine, 1, MPI_INT, MPI_COMM_WORLD);
  MPI_Unpack(task_buffer, task_buffer_size, &position, node_map, N, MPI_INT, MPI_COMM_WORLD);

  return status.MPI_SOURCE;
}

int sendResponse(Response& msg, int N) {
  int position = 0; //current position in the buffer

  MPI_Pack(&msg.task_number, 1, MPI_INT, reply_buffer, reply_buffer_size, &position, MPI_COMM_WORLD);
  MPI_Pack(&msg.s, 1, MPI_INT, reply_buffer, reply_buffer_size, &position, MPI_COMM_WORLD);
  MPI_Pack(&msg.t, 1, MPI_INT, reply_buffer, reply_buffer_size, &position, MPI_COMM_WORLD);
  MPI_Pack(&msg.graph_size, 1, MPI_INT, reply_buffer, reply_buffer_size, &position, MPI_COMM_WORLD);

  MPI_Pack(&msg.flow, 1, MPI_DOUBLE, reply_buffer, reply_buffer_size, &position, MPI_COMM_WORLD);
  MPI_Pack(&msg.time_per_mincut, 1, MPI_DOUBLE, reply_buffer, reply_buffer_size, &position, MPI_COMM_WORLD);
  MPI_Pack(&msg.time_to_contract, 1, MPI_DOUBLE, reply_buffer, reply_buffer_size, &position, MPI_COMM_WORLD);
  MPI_Pack(&msg.num_of_cuts, 1, MPI_INT, reply_buffer, reply_buffer_size, &position, MPI_COMM_WORLD);
  MPI_Pack(&msg.balance_factor, 1, MPI_FLOAT, reply_buffer, reply_buffer_size, &position, MPI_COMM_WORLD);

  MPI_Pack(msg.cut, N, MPI_INT, reply_buffer, reply_buffer_size, &position, MPI_COMM_WORLD);

  int to = 0; //always to master
  int tag = 1;
  int r = MPI_Send(reply_buffer, position, MPI_PACKED, to, tag, MPI_COMM_WORLD);

  return r;
}

void init_task_response_buffers(int N)
{
  // memory allocation 
  task_buffer_size = sizeof(Task) + N*sizeof(int);
  task_buffer = (int *)calloc(1, task_buffer_size); 

  reply_buffer_size = sizeof(Response) + N*sizeof(int);
  reply_buffer = (int *)calloc(1, reply_buffer_size);
}


void release_task_response_buffers()
{
  free(task_buffer);
  free(reply_buffer);
}
