#pragma once

#include <iostream>
#include <set>

/* Structure used to send and receive msgs*/
class Response {
 public:
  int task_number;
  int s, t;
  int graph_size;
  double flow; /* max flow */
  double time_per_mincut;
  double time_to_contract;
  int num_of_cuts;
  float balance_factor; // measure of how balanced is the cut
  int *cut;    /* list of vertices on s side */
  // WARNING: if you change the fields, check the reply_buffer allocation size and MPI pack/unpack

  Response(int n);
  ~Response();
};

struct Task {
  int task_number;
  int s, t;
  int pivot_size;
  int part_len; // partition size or the pivot degree  (-1 indicates no contraction)
  int refine;  // refine the contracted graph or start again from the input graph
  // WARNING: if you add new fields, check the task_buffer allocation size

  Task();
  Task (int task_number, int s, int t, int pivot_size, int part_len, int refine);
  ~Task();
};

struct Slave_state {
//  struct Task task;
  std::set<int> pivot;
};

void init_task_response_buffers(int N);
void release_task_response_buffers();
void print_node_map(int *node_map, int n, std::ostream& os);
int sendTask(Task& task, int dest, int N, int* node_map);
void print_task(Task& task, int* node_map, int N);
void print_reply (Response& msg, int N);
int receiveResponse(Response& msg, int N);
int receiveTask(Task& task, int N, int *node_map);
int sendResponse(Response& msg, int N);

