#pragma once

#include <iostream>
#include <set>
#include <vector>

#include "edgelist.hpp"
#include "task.hpp"

using namespace std;

struct GHTreeNode;


struct GHTreeEdge {
  double capacity;       // edge capacity
  GHTreeNode *tail, *head; // nodes of the edge
  int tail_vertex, head_vertex; // pair of vertices that produced the min-st-cut
                                // that originated this edge
  GHTreeEdge *next; // next edges on tail's list 
  GHTreeEdge *reverse;  // reverse edge
};

struct GHTreeNode {
  GHTreeEdge *first_edge;
  int degree;
  int num_vertices;      // number of vertices in the node 
  int id;                // vertex id when |vertices|=1 
  set<int> vertices;    // graph vertices in the tree node 
  GHTreeNode* next_pivot; // next node with more than one vertex
  GHTreeNode* prev_pivot; // previous node with more than one vertex
  set<int>::iterator next_source, next_sink; // used to enumerate s-t pairs to produce tasks
  int st_step, st_disp, st_index; // used to enumerate s-t pairs to produce tasks
  bool st_sink_reset;
  bool done_every_pair; // if all pairs had been enumerated, then skip the node. 
};

class GHTree {
  int num_nodes, max_num_nodes;        //  current number of nodes / max number of nodes
  GHTreeNode *nodes;   // all nodes of the tree
  GHTreeEdge *edges;   // all edges of the tree 
  GHTreeEdge *nextFreeEdge;
  // int source, sink;
  int partition_size;
  Edgelist_vector *input_graph; // input graph
  GHTreeNode& next_free_node ();
  void scan_subtree(const GHTreeNode* u, const GHTreeNode* v, int node_number);  
  vector<GHTreeNode*> vertex_node_map;
  GHTreeNode *first_pivot;
  GHTreeNode *current_pivot;

  set<int>::iterator it_last(set<int>& S)  {  return --S.end(); }
  set<int>::iterator it_plus_one(set<int>::iterator it) {  set<int>::iterator next = it;  return ++next; }
  bool inc_source(GHTreeNode* pivot);
  bool inc_sink(GHTreeNode* pivot);
  void reset_task_iterators(GHTreeNode& v);
 public:
  int* map_back; // maps nodes from G_contracted to the input graph/gh tree
  int* node_map; // maps nodes from input graph to G_contracted
  
  GHTree(Edgelist_vector& input_graph);
  ~GHTree();
  int size() { return num_nodes; }
  void add_edge(GHTreeNode& u, GHTreeNode& v, double capacity, int s, int t);
  int build_node_map(const GHTreeNode* pivot);
  int expand_tree(Response& msg);
  friend ostream& operator<<(ostream& os, GHTree& T);
  bool in_node(int vertex, const GHTreeNode &node);
  GHTreeNode* pick_pivot();
  GHTreeNode* pick_pivot_2();
  GHTreeNode* pick_pivot_3(set<int>& last_pivot, int& refine);
  void pick_source_and_sink(int& source, int& sink, GHTreeNode* pivot);
  void pick_source_and_sink_2(int& source, int& sink, GHTreeNode* pivot);
  void pick_source_and_sink_3(int& source, int& sink, GHTreeNode* pivot);
  void print  (bool print_vertices, bool first_node);
  int get_partition_size() { return partition_size; };
  void to_dot();
  void circ_inc_pivot (GHTreeNode*& n);
  bool failed_task(Response& msg);
};

void build_contracted_graph(Edgelist_vector& input_graph, Edgelist& G_contracted, int* node_map, int N_contracted);
void build_contracted_graph_from_contracted(Edgelist& old_graph, Edgelist& G_contracted, int* node_map, int N_contracted, int* node_id);

  /* not needed (yet)
typedef struct st_contracted_to_tree {
  char contracted;   "1 if node is a contracted node, 0 otherwise "
  int node;          "either the index of a tree node or of the input graph" 
} * contracted_to_tree; */


/* node partition will be a vector: part[i] = -1 if i should not be contracted or */
/* part[i] = j if i belongs to node partition j */
